#ifndef CAMERA_H
#define CAMERA_H

#include "vector.h"
#include "quat.h"
#include "geometry.h"


class M_Camera {
public:
	//creation
	M_Camera();
	M_Camera(M_Camera *src);	//create a cam from another cam
	bool Init();

	//destruction
	~M_Camera() {Kill();}
	void Kill();

	//OpenGL transformation
public:

#define CAMERA_RESET_PROJECTION		1
#define CAMERA_RESET_ROTATION		2
#define CAMERA_RESET_TRANSLATION	4
#define CAMERA_RESET_SCALE			8

	void TransformScene(int flags);

	//the resetting of the frustum
	void CalculateFrustum(void);

	void EnableClipPlane(Plane *plane);
	void DisableClipPlane(void);
private:
	void SetProjection(void);
	void SetModelview(int flags);
	//the functions of SetModelview() - rotate, translate, scale (scale used for mirrors)
	void RotateScene(void);
	void TranslateScene(void);
	void ScaleScene(void);

public:
	void Activate(void);
	bool active;

// ********** FRUSTUM SHAPE ********** 
public:
	float	FOV;	//not used quite yet
	float	znear, zfar;
	Vector	axis[3];
	
	//created based on znear, zfar, FOV, position, and angle
	Vector	frustum_vtx[8];
	//six planes of occulsion
	Plane	frustum_plane[6];

	//created based upon frustum coordinates
	float	xmin, xmax, ymin, ymax;	//the quad containing the frustum projected on the XY plane HORRIBLE HORRIBLE HACK - get rid of this

// ********** ORIENTATION ********** 
public:
	Vector	position;
	Quat	angle;
	Vector	scale;

private:	//used for mirros - for the time being
	Plane	clip_plane;
	bool	clip_plane_active;


// ********** TRACKING ********** 
#define CAM_TRACK_NO		0
#define CAM_TRACK_YES		1
#define CAM_TRACK_CHASE		2

public:
	void Update(void);

	void SetTrackPosition(Vector *v) {trackPosition = v;}
	void SetTrackAngle(Quat *q) {trackAngle = q;}
	void SetTrackMethod(int m) {trackMethod = m;}
	void SetTrackAngleSpeed(float c) {trackSlerpC = c;}
	void SetTrackPosSpeed(float c) {trackLerpC = c;}
	void SetChaseOffset(Vector v) {chaseOffset = v;}

private:
	Vector	*trackPosition;
	Quat	*trackAngle;
	int		trackMethod;
	float	trackLerpC;		//the LERP coefficient - tracking position
	float	trackSlerpC;	//the slerp coefficient from the current to the tracking angle
	Vector	chaseOffset;

};

#endif