#include <windows.h>	//for NULL definition
#include "chain.h"

Chain::Chain()
{
	chain_parent = NULL;
	chain_prev = NULL;
	chain_next = NULL;
}

Chain::~Chain()
{
	//if this is attached to a chain (if it has a parent chain)
	if (chain_parent)
		Unlink();		//then unlink it
	else	//if not, check to see if it has children, and unlink them all
		UnlinkAll();
}

//methods called from links within a chain
//to add them to a parent chain

void Chain::Link(Chain *p)
{
	//if a parent already exists then unlink it from its current chain
	if (chain_parent) Unlink();

	chain_parent = p;
	if (!chain_parent) return;

	if (!chain_parent->chain_head) chain_parent->chain_head = this;
	if (chain_parent->chain_tail) chain_parent->chain_tail->chain_next = this;

	chain_prev = chain_parent->chain_tail;
	chain_next = NULL;
	chain_parent->chain_tail = this;
}

#include <assert.h>

//adds link to the specified chain's parent, just after the specified chain
void Chain::LinkBefore(Chain *c)
{
	//unlink it from its current parent (if any)
	if (chain_parent) Unlink();

	assert(c);
	assert(c->chain_parent);

	//set its new parent
	chain_parent = c->chain_parent;
	
	//set the current link prev and next
	chain_prev = c->chain_prev;
	chain_next = c;

	//link all other to this one
	c->chain_prev = this;
	//if there exists a prev link in the chain then set its chain_next to 'this'
	if (chain_prev) chain_prev->chain_next = this;
	//else, 'this' lies at the beginning of the chain - set the parent head to 'this'
	else chain_parent->chain_head = this;
}

//adds link to the specified chain's parent, just after the specified chain
void Chain::LinkAfter(Chain *c)
{
	//unlink it from its current parent (if any)
	if (chain_parent) Unlink();

	assert(c);
	assert(c->chain_parent);

	//set its new parent
	chain_parent = c->chain_parent;
	
	//set the current link prev and next
	chain_prev = c;
	chain_next = c->chain_next;

	//link all other to this one
	c->chain_next = this;
	//if there exists a next link in the chain then set its prev to 'this'
	if (chain_next) chain_next->chain_prev = this;
	//else, 'this' lies at the end of the chain - set the parent tail to 'this'
	else chain_parent->chain_tail = this;
}

void Chain::Unlink(void)
{
	if (!chain_parent) return;
	
	if (this == chain_parent->chain_head) chain_parent->chain_head = chain_next;
	if (this == chain_parent->chain_tail) chain_parent->chain_tail = chain_prev;

	if (chain_prev) chain_prev->chain_next = chain_next;
	if (chain_next) chain_next->chain_prev = chain_prev;

	chain_parent = NULL;
}

//methods called by a parent chain
//to remove children
// NOTICE - if this is called not by a link within a chain
//but instead by a child of a certain chain
//then this will unlink all members of that chain linked *AFTER* 'this'
void Chain::UnlinkAll(void)
{
	Chain *v, *next;
	for (v = chain_head; v; )
	{
		next = v->chain_next;
		if (v->chain_parent == this) v->Unlink();
		v = next;
	}
	chain_head = chain_tail = NULL;
}

void Chain::DeleteAll(void)	//deletes all children of the chain
{
	Chain *v, *next;
	for (v = chain_head; v; )
	{
		next = v->chain_next;
		if (v->chain_parent == this) delete v;
		v = next;
	}
	chain_head = chain_tail = NULL;
}

int Chain::Count(void)
{
	int cnt = 0;
	for (Chain *v = chain_head; v; v=v->chain_next) cnt++;
	return cnt;
}