#ifndef CHAIN_H
#define CHAIN_H

/*
  one class to contain the header/tail of the link
  another class to contain the prev/next pointers, as well as the pointer to the data
*/

class Chain {

public:
	Chain();				//constructor : nulls all pointers
	~Chain();				//destructor - unlinks all chains

	//child chain calls
	void Link(Chain *p);	//adds the link to the end of the chain
	void LinkBefore(Chain *c);	//adds link to the specified chain's parent, just before the specified chain;
	void LinkAfter(Chain *c);	//adds link to the specified chain's parent, just after the specified chain
	void Unlink(void);		//and removes the link from the current parent chain
	//parent chain calls
	void UnlinkAll(void);	//unlinks all children of the chain
	void DeleteAll(void);	//deletes all children of the chain

	Chain	*ChainPrev(void) {return chain_prev;}	//allow reading of these variables
	Chain	*ChainTail(void) {return chain_tail;}	//but not writing
	Chain	*ChainNext(void) {return chain_next;}
	Chain	*ChainHead(void) {return chain_head;}

	int Count(void);

private:

	//union up the head/tail - so the Link class can be used as a chain link
	//as well as a list container

	union {
		Chain	*chain_prev;		//points to the prev node of the link within the chain
		Chain	*chain_tail;		//points to the tail of the chain
	};
	union {
		Chain	*chain_next;		//points to the next node in the chain
		Chain	*chain_head;		//points to the head of the chain
	};

	//pointer to the chain parent (which contains the head/tail of the chain)
	Chain		*chain_parent;
};

//parallel-linked chains.  link these to elements in arrays.  sort them accordingly without
//having to sort the contents of the data
template <class DataClass> class ChainPll : public Chain {
public:
	DataClass	*chain_data;
};

#endif