#include <windows.h>
#include <stdio.h>

#include "console.h"
#include "error.h"

#include "main.h"

// ***********************************
// M_CVar
// ***********************************

Chain M_CVar::parentChain;

M_CVar *M_CVar::MatchString(char *str)
{
	M_CVar *v = (M_CVar *)M_CVar::parentChain.ChainHead();
	for (; ; )
	{
		if (_stricmp(v->name, str))
		{
			v = (M_CVar *)v->ChainNext();
			if (!v) break;
			continue;
		}
		break;
	}
	return v;
}

	//echoes its data to the string in dest - specific per-type
void M_CVar::EchoData(char *dest)
{
	if (!data) return;	//duh

	switch (type)
	{
	case CVAR_INT:		sprintf(dest, "%d", *(int *)data); break;
	case CVAR_FLOAT:	sprintf(dest, "%f", *(float *)data); break;
	case CVAR_VECTOR:	sprintf(dest, "%f %f %f", ((Vector *)data)->x, ((Vector *)data)->y, ((Vector *)data)->z);	break;
	case CVAR_QUAT:		sprintf(dest, "%f %f %f %f", ((Quat *)data)->x, ((Quat *)data)->y, ((Quat *)data)->z, ((Quat *)data)->w);	break;
	}
}

	//pointer to an array of pointers to strings to contain the value
void M_CVar::SetValue(char **argv)
{
	switch (type)
	{
	case CVAR_INT:
		if (!argv[1]) break;		*(int *)data = atoi(argv[1]);
		break;
	case CVAR_FLOAT:
		if (!argv[1]) break;		*(float *)data = (float)atof(argv[1]);
		break;
	case CVAR_VECTOR:
		if (!argv[1]) break;		((Vector *)data)->x = (float)atof(argv[1]);
		if (!argv[2]) break;		((Vector *)data)->y = (float)atof(argv[2]);
		if (!argv[3]) break;		((Vector *)data)->z = (float)atof(argv[3]);
		break;
	case CVAR_QUAT:
		if (!argv[1]) break;		((Quat *)data)->x = (float)atof(argv[1]);
		if (!argv[2]) break;		((Quat *)data)->y = (float)atof(argv[2]);
		if (!argv[3]) break;		((Quat *)data)->z = (float)atof(argv[3]);
		if (!argv[4]) break;		((Quat *)data)->w = (float)atof(argv[4]);
		break;
	}
}

void M_CVar::ExecFunc(int argc, char **argv)
{
	if (func) func(argc, argv, param);
}


void M_CVar::ListCmds(void) { for (M_CVar *v = (M_CVar *)M_CVar::parentChain.ChainHead(); v; v=(M_CVar *)v->ChainNext()) main_console->Print("%s\n", v->name); }
void M_CVar_ListCmds(int argc, char **argv, void *param) {M_CVar::ListCmds();}


// ***********************************
// M_Console
// ***********************************

#include <gl/gl.h>
#include "random.h"

//construction / creation
M_Console::M_Console()
{
	memset(argv,0,sizeof(argv));
	argc = 0;
	fontList = 0;

	ClearInput();
	memset(textbuffer, 0, sizeof(textbuffer));

	isActive = false;
}

#define CONSOLE_TYPE_COUNT 4
int ConsoleDropTime = 300;	//MS to drop console (change to frames?)
float ConsoleTimeFrac;
int ConsoleType;

bool M_Console::Init(int w, int h)
{
	width = w;
	height = h;
	dropdown = h-1;

	font = new Texture();
	if (!font) return false;
	if (!font->Create("WorldData/font.bmp")) return false;

	fontList = glGenLists(256);
	if (!fontList)  return false;
	for (int y = 0; y < 16; y++)
	{
		for (int x = 0; x < 16; x++)
		{
			glNewList(fontList+x+y*16, GL_COMPILE);

			glBindTexture(GL_TEXTURE_2D, font->texture_id);
			glBegin(GL_QUADS);
			glTexCoord2f((float)x/16.f, (float)(y)/16.f);glVertex2f(0,0);
			glTexCoord2f((float)(x+1)/16.f, (float)(y)/16.f);glVertex2f(1,0);
			glTexCoord2f((float)(x+1)/16.f, (float)(y+1)/16.f);glVertex2f(1,1);
			glTexCoord2f((float)x/16.f, (float)(y+1)/16.f);glVertex2f(0,1);
			glEnd();
			glTranslatef(1,0,0);

			glEndList();
		}
	}


	new M_CVar((void *)NULL, "ListCmds", M_CVar_ListCmds, NULL);
	new M_CVar(&ConsoleDropTime, "ConsoleDropTime");

	return true;
}

#include "keyboard.h"

bool M_Console::KeyDown(int key)
{
	if (key == VkKeyScan('`'))
	{
		Toggle();

		if (isActive)
		{
			memset(inputcmd, 0, sizeof(inputcmd));
			inkey_next = 0;
			ConsoleType = MRand() % CONSOLE_TYPE_COUNT;
			dropstart = timeGetTime();
		}
		else
		{
			ConsoleType = 0;
		}

		return true;	//true means we interpreted the key
	}

	if (isActive)
	{
		if (key == VK_RETURN)
		{
			//parse the current command buffer
			ExecCommand(inputcmd);
			ClearInput();
			return true;
		}
		else if (key == VK_SPACE)
		{
			AddToInput(' ');
			return true;
		}
		else if (key == VK_BACK)
		{
			AddToInput(8);
			return true;
		}
		else
		{
			if (main_keyboard && main_keyboard->keystate[VK_SHIFT])
				key |= 0x100;

			//replace this with a key mapping function
			for (int ch = 32; ch < 128; ch++)
			{
				if (key == VkKeyScan(ch))
				{
					AddToInput((char)ch);
					return true;
				}
			}
		}
	}

	return false;
}

void M_Console::AddToInput(char key)
{
	//backspace
	if (key == 8)
	{
		if (inkey_next > 0)
		{
			inkey_next--;
			inputcmd[inkey_next] = 0;
		}
		return;
	}

	//any other key
	if (inkey_next >= sizeof(inputcmd)) return;
	inputcmd[inkey_next] = key;
	inkey_next++;
}
void M_Console::ClearInput(void)
{
	memset(inputcmd, 0, sizeof(inputcmd));
	inkey_next = 0;
}

void M_Console::Toggle(void)
{
	if (!isActive)	isActive = true;
	else			isActive = false;
}


// output

//and the per-object function
void M_Console::Print(const char *_format, ...)
{
	char text[1024];
	va_list ap;

	va_start(ap, _format);
	vsprintf(text, _format, ap);
	va_end(ap);

	AddString(text);
}

void M_Console::AddString(char *append_data)
{
	int append_data_len = strlen(append_data);

	if (!append_data) return;
	if (!*append_data) return;

	int buffer_data_len = strlen(textbuffer);
	if (buffer_data_len + append_data_len > sizeof(textbuffer)-1)
	{
		memcpy(	textbuffer,
				&textbuffer[append_data_len],
				buffer_data_len - append_data_len);

		textbuffer[buffer_data_len - append_data_len] = 0;
	}
	strcat(textbuffer, append_data);
}

void M_Console::ExecCommand(char *cmd)
{
	//parse the arguments of the command
	ParseArguments(cmd);
	
	//if there are none then return
	if (argc == 0) return;

	//match the first argument to a known variable
	M_CVar *var = M_CVar::MatchString(argv[0]);
	if (var == NULL)
	{
		Print(" - unknown cvar\n");
		return;
	}

	if (argc > 1) var->SetValue(argv);
	var->ExecFunc(argc, argv);

	char dest_str[256];
	var->EchoData(dest_str);
	Print("%s\n", dest_str);
}

void M_Console::ParseArguments(char *cmd)
{
	memset(argv,0,sizeof(argv));
	argc = 0;

	//cycle to the beginning of some text (past the spaces)
	for (; *cmd == 32; cmd++) if (*cmd == 0) return;
	//if we hit a null terminator then return - no arguments detected

	argv[0] = argv[1] = cmd;		//set the first argument
									//and set the second argument to begin searching at the first loc
	argc = 1;
	do {
		argv[argc]++;		//and search for the end of it

		//string terminator
		if (*argv[argc] == 0) break;

		//space - argument separator
		if (*argv[argc] == 32)
		{
			//if we find a space then set it to a null terminator
			*argv[argc] = 0;
			//increment the pointer to one past the new terminator
			argv[argc]++;
			//search out the end of the spaces
			for (; *argv[argc] == 32; argv[argc]++) if (*argv[argc] == 0) break;

			//so that by here, argv[argc] is a pointer to the beginning of the new argument

			//increase the argument counter
			argc++;
			//if we are full on arguments - leave
			if (argc == 16) break;
			//set the new argument pointer to the current argument pointer position
			argv[argc] = argv[argc-1];
			//and continue through the loop
		}
		
	} while (1);
}

//display
void M_Console::Display(void)
{
	int current_dropdown = dropdown;
	if (!isActive) current_dropdown = 4;

	//0 - console just dropped
	//1 - console set in position
	ConsoleTimeFrac = (float)(timeGetTime() - dropstart) / (float)ConsoleDropTime;
	if (ConsoleTimeFrac > 1) ConsoleTimeFrac = 1;
	if (ConsoleTimeFrac < 0) ConsoleTimeFrac = 0;

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0,width,height,0,-1000,1000);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();


	//distort this according to time_frac
	switch (ConsoleType)
	{
	case 1:
		glTranslatef((float)width/2.f, (float)height/2.f, 0);
		glScalef(ConsoleTimeFrac, ConsoleTimeFrac, 1.f);
		glRotatef(1000 - 1000 * ConsoleTimeFrac, 0,0,1);
		glTranslatef(-(float)width/2.f, -(float)height/2.f, 0);
		break;
	case 2:
		{
			if (ConsoleTimeFrac < 0.5)
				glScalef(.2f, 2*ConsoleTimeFrac, 1.f);
			else
				glScalef(1.6f*(ConsoleTimeFrac-.5f) + .2f, 1.f, 1.f);
		}
		break;
	}

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glEnable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	glEnable(GL_BLEND);

	if (isActive)
	{
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glBindTexture(GL_TEXTURE_2D, 0);
		glColor4f(0,0,0,.75f);
		glBegin(GL_QUADS);
		glVertex2f(0,0);
		glVertex2f((float)width,0);
		glVertex2f((float)width,(float)height);
		glVertex2f(0,(float)height);
		glEnd();
	}

	glBlendFunc(GL_ONE, GL_ONE);

	glColor3f(1,1,1);

	int lastListBase;
	glGetIntegerv(GL_LIST_BASE, &lastListBase);
	glListBase(fontList-32);

	//print the current cmd being typed
	glTranslatef(0,(float)current_dropdown,0);
	
	DisplayLine(inputcmd, strlen(inputcmd));

	//print the text in the console
	char *start, *end;
	for (end = textbuffer; *end; end++) if (end >= textbuffer + sizeof(textbuffer)) break;
	do {
		for (start = end; *start >= 32 && start >= textbuffer; start--);
		glTranslatef(0,-1,0);
		DisplayLine(&start[1], end-start);
		for (end = start-1; *end < 32 && end >= textbuffer; end--);
	} while (end >= textbuffer);

	glListBase(lastListBase);

	glPopAttrib();

	glPopMatrix();		//pop the modelview matrix
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();		//pop the projection matrix

	glMatrixMode(GL_MODELVIEW);	//and set the current matrix back to modelview
}

#include <math.h>
#include "const.h"
void M_Console::DisplayLine(char *p, int len)
{
	glPushMatrix();

	switch (ConsoleType)
	{
	//all transformations done via OpenGL matrix manipulation
	case 0:
	case 1:
	case 2:
		glCallLists(len, GL_UNSIGNED_BYTE, p);
		break;
	//we need to tweak the verticies!
	//save me, oh save me, my NVidia Vertex Shaders!
	default:
		for (int a = 0; a < len; a++)
		{
			int tx = (*(unsigned char *)p % 16);
			int ty = (*(unsigned char *)p / 16) - 2;

			float vtx[4][2] = {
				{(float)a,		0},
				{(float)(a+1),	0},
				{(float)(a+1),	1},
				{(float)a,		1}};

			switch (ConsoleType)
			{
			case 3:
				{
					//float scalar = sin(ConsoleTimeFrac * M_PI*.5);
					for (int v = 0; v < 4; v++)
					{
//						float scalar = ConsoleTimeFrac;
//						vtx[v][0] *= scalar;
						if ((vtx[v][0] / (float)width) < (1.f - ConsoleTimeFrac)) vtx[v][0] = 0;
					}
				}
				break;
			}

			glBindTexture(GL_TEXTURE_2D, font->texture_id);
			glBegin(GL_QUADS);
			glTexCoord2f((float)tx/16.f, (float)(ty)/16.f);			glVertex2fv(vtx[0]);
			glTexCoord2f((float)(tx+1)/16.f, (float)(ty)/16.f);		glVertex2fv(vtx[1]);
			glTexCoord2f((float)(tx+1)/16.f, (float)(ty+1)/16.f);	glVertex2fv(vtx[2]);
			glTexCoord2f((float)tx/16.f, (float)(ty+1)/16.f);		glVertex2fv(vtx[3]);
			glEnd();

			p++;
		}
		break;
	}

	glPopMatrix();
}

//destruction

void M_Console::Kill()
{
	delete font;
	M_CVar::parentChain.DeleteAll();
}