#ifndef CONSOLE_H
#define CONSOLE_H

#include "quat.h"
#include "vector.h"
#include "chain.h"

#define CVAR_VOIDP	0
#define CVAR_INT	1 
#define CVAR_FLOAT	2
#define CVAR_VECTOR	3
#define CVAR_QUAT	4

/*
This is the console variable class.
M_CVars are attached to certain variables 
M_CVars should only be attached to variables which can be dynamically changed without any other adjustments needed

  the method I am using:
  we pass the following information to the console variable:
  data - the variable pointer
  type - tye type of variable (for converting later on)
  name - the name of the data - used for interpreting console commands
*/
class M_CVar : public Chain {
public:
	M_CVar(void *p,		char *n, void (*f)(int, char **, void *) = NULL, void *o = NULL)	{data = p; name = n; func = f; param = o; type = CVAR_VOIDP;	Link(&M_CVar::parentChain);}
	M_CVar(int *p,		char *n, void (*f)(int, char **, void *) = NULL, void *o = NULL)	{data = p; name = n; func = f; param = o; type = CVAR_INT;		Link(&M_CVar::parentChain);}
	M_CVar(float *p,	char *n, void (*f)(int, char **, void *) = NULL, void *o = NULL)	{data = p; name = n; func = f; param = o; type = CVAR_FLOAT;	Link(&M_CVar::parentChain);}
	M_CVar(Vector *p,	char *n, void (*f)(int, char **, void *) = NULL, void *o = NULL)	{data = p; name = n; func = f; param = o; type = CVAR_VECTOR;	Link(&M_CVar::parentChain);}
	M_CVar(Quat *p,		char *n, void (*f)(int, char **, void *) = NULL, void *o = NULL)	{data = p; name = n; func = f; param = o; type = CVAR_QUAT;		Link(&M_CVar::parentChain);}

	void Reset(void *p)		{data = p; type = CVAR_VOIDP;}
	void Reset(int *p)		{data = p; type = CVAR_INT;}
	void Reset(float *p)	{data = p; type = CVAR_FLOAT;}
	void Reset(Vector *p)	{data = p; type = CVAR_VECTOR;}
	void Reset(Quat *p)		{data = p; type = CVAR_QUAT;}

	~M_CVar() {Unlink();}	//this is a definate - unlink a CVar upon destruction


	void EchoData(char *dest);	//echoes its data to the string in dest - specific per-type
	void SetValue(char **argv);	//pointer to an array of pointers to strings to contain the value
	void ExecFunc(int argc, char **argv);

//i could group these together into a CVar List class
	static Chain parentChain;
	static M_CVar *MatchString(char *str);	//console function to find a matching cvar string
	static void ListCmds(void);	//this is the console command to pump all console commands to the screen

private:

	void	*data;	//the actual data
	int		type;	//the type of data
	char	*name;	//the name of the functino

	void	(*func)(int, char **, void *);	//the update function (argc, argv, and the param)
	void	*param;					//the object which owns the update function
};


/*
The Console Class!
This holds the font, and handles all the IO commands
*/
#include "texture.h"

class M_Console {
public:
	M_Console();
	bool Init(int w, int h);

	~M_Console() {Kill();}
	void Kill();

	bool KeyDown(int key);

	//shell for AddString() - interprets the *_format
	void Print(const char *_format, ...);

	//reads the string from the input window
	void ReadCommand(void);

	//displays the console
	void Display(void);
	void DisplayLine(char *p, int len);

	//interprets a command
	void ExecCommand(char *cmd);

private:
	void Toggle(void);
	void AddToInput(char key);
	void ClearInput(void);

	int inkey_next;
	char inputcmd[256];
	char textbuffer[2048];


	//adds a string to the output window
	void AddString(char *append_data);

	//interprets a command
	void ParseArguments(char *cmd);

	char *argv[16];				//maximum 16 arguments
	int argc;

	Texture *font;		//texture class of the font used
	int fontList;		//first ID of the font list

	//used for toggling the console
	bool isActive;

	//amount of chars to display in the console
	int width, height, dropdown;
	int dropstart;
};


#endif