#ifndef ENTITY_H
#define ENTITY_H

#include "vector.h"
#include "quat.h"
#include "chain.h"

#include "model.h"
#include "cmodel.h"
#include "texture.h"

#include "geometry.h"

#include "types.h"

//#define DEBUG_CMODEL

class Entity : public Chain {

// ********** CONSTRUCTION/DESTRUCTION ********** 
public:
	Entity() {Init();}			//constructs an entity	- inits it and links it to the list
	~Entity() {Kill();};			//destructs an entity	- removes it from the list

	void Init(void);
	void Kill(void);

// ********** ENTITY VARIABLES ********** 
public:
	bool active;
	bool check_collision;

	char	type[64];

	Vector	position, velocity;
	Quat	angle, rotation;

	//the base bounding box - used for orientating/calcing maxbbox for quick collision detection
	//also used for quick hack collision model creation when one is not provided
	Box		bbox;
	//bounding box of the rotated/translated collision model - do we need this?  yes - saves on calculations
	//this is tested during collision testing - calculated during Ent_AlignBBox()
	Box		maxbbox;

	//useful to have around - the entity matrix
	Vector	axis[3];

	bool	transparent;
	color_t	color;
	Texture	*texture;

	float	scale;
	Model	*model;		//display model - triangle mesh
	CModel	*c_model;	//collision model - convex hull - polygonal surface 

	void	(*Think)();
	void	(*Touch)();

	bool touched;
	double timer;

#ifdef DEBUG_CMODEL
	int		d_cmodel_vtxCount;
	Vector	d_cmodel_vtx[100];
	int		d_cmodel_planeCount;
	Plane	d_cmodel_plane[100];
	int		d_cmodel_edgeCount;
	int		*d_cmodel_edge;
#endif


// ********** ENTITY FUNCTIONS **********  called by Engine
	//the per-entity display function
public:
	void Display(void);
};

//used by engine.cpp - for main construction/destruction
bool InitEntities(void);
void UpdateEntities(void);
void DisplayEntities(void);
void ShutdownEntities(void);

//used by engine.cpp - because engine.cpp is a misfit
void Ent_FireShot(Entity *e);
void StartPlayer(Entity *e);
void StartRock(Entity *e);
Entity *CreateEntity(void);

//used by world.cpp - for world rendering
#define WorldSize	2500

//used by our entity.cpp friend: collision.cpp
#define MAX_ENTS	5000
bool EntIsActive(Entity *e);
extern Entity *main_ent, *client_ent, *touch_ent, *entityarray;

#endif