#ifndef PLANE_H
#define PLANE_H

#include "vector.h"

class Line;

class Plane {
public:

	Vector pos;
	Vector normal;

	//creates a plane equation based upon the passed position and normal vectors
	void CalculateEquation(double *output);

	//returns the distance from the vector v to the plane
	float Distance(Vector *v);

	//calculates the point on the plane closest to 'in' and stores result in 'out'
	void ClosestPoint(Vector *in, Vector *out);

	//calculates the position of where a line will intersect this plane
	//store the intersection vector in vout
	//returns the fraction of the line
	float LineIntersect(Line *lin, Vector *vout);

	//calculates the fraction of the line's dir from its position to the intersection of the plane
	float LineIntersectFraction(Line *lin);
};

class Line {
public:

	Vector pos;
	Vector dir;

	//fills the line with the line intersection of the two planes
	void PlaneIntersect(Plane *p1, Plane *p2);
	//if the two planes do not intersect then the line dir will be equal to zero
};

class Box {
public:
	struct {
		Vector min,max;
	};

	Vector *vec(void) {return (Vector *)(&min);}

	void Stretch(Vector *s_min, Vector *s_max);
	void Stretch(Box *box)	{	Stretch(&box->min, &box->max);	}
	void Stretch(Vector *v) {	Stretch(v,v);	}

	float Length(void);
	Vector Center(void);
	bool Contains(Vector *v);
	bool Contains(Box *b);
	bool Intersects(Box *b);

	bool CheckLineIntersect(Vector *src, Vector *delta);
	bool CheckLineIntersect(Line *l) { return CheckLineIntersect(&l->pos, &l->dir);}
	float CheckBoxIntersect(Box *b, Vector del, Vector *res);

	void CalcVtxs(Vector *vtx)
	{
		vtx[0] = Vector(min.x, min.y, min.z);
		vtx[1] = Vector(min.x, min.y, max.z);
		vtx[2] = Vector(min.x, max.y, min.z);
		vtx[3] = Vector(min.x, max.y, max.z);
		vtx[4] = Vector(max.x, min.y, min.z);
		vtx[5] = Vector(max.x, min.y, max.z);
		vtx[6] = Vector(max.x, max.y, min.z);
		vtx[7] = Vector(max.x, max.y, max.z);
	}

	void CalcPlanes(Plane *p)
	{
		p[0].pos = min;	p[0].normal = Vector(-1,0,0);
		p[1].pos = min;	p[1].normal = Vector(0,-1,0);
		p[2].pos = min;	p[2].normal = Vector(0,0,-1);
		p[3].pos = min;	p[3].normal = Vector(1,0,0);
		p[4].pos = min;	p[4].normal = Vector(0,1,0);
		p[5].pos = min;	p[5].normal = Vector(0,0,1);
	}
};

#endif