#include <windows.h>
#include <gl/gl.h>

#include "mirror.h"

#define M_STENCIL_MIRROR_BITS 0xF	//the low 4 bits of the byte (right now all we're using of the stencil buffer)

void M_Mirror::Render(void)
{
	//draw our reflection quad
	glBegin(GL_QUADS);
	glVertex3fv(v[0].v);
	glVertex3fv(v[1].v);
	glVertex3fv(v[2].v);
	glVertex3fv(v[3].v);
	glEnd();
}

void M_Mirror::BeginRenderStencil(void)
{
//render reflective trees to stencil buffer

	//disable the depth write mask and color mask
	glDepthMask(GL_FALSE);
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);

	//set up the stencil buffer
	glEnable(GL_STENCIL_TEST);
//	glClear(GL_STENCIL_BUFFER_BIT);
	glStencilOp(GL_KEEP  , GL_ZERO , GL_REPLACE);
	glStencilFunc(GL_ALWAYS, 1, M_STENCIL_MIRROR_BITS);

	Render();

	//disable color mask and depth test
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	glDepthMask(GL_TRUE);

	//only render where stencil buffer is 1
	glStencilFunc(GL_EQUAL, 1, M_STENCIL_MIRROR_BITS);
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

	//clear the depth buffer and render whats behind the mirror
	glClear(GL_DEPTH_BUFFER_BIT);

	//flip the faces - because its mirrored	(make me dependant on the camera's scale matrix?)
	glCullFace(GL_FRONT);

}

void M_Mirror::EndRenderStencil(void)
{
	glCullFace(GL_BACK);

	//disable depth testing - solely rely on stencil testing
	//why so?  because depth testing produces inaccuracies
	//where polygons in the reflection image intercept the mirror
	glDisable(GL_DEPTH_TEST);

	glStencilFunc(GL_EQUAL, 1, M_STENCIL_MIRROR_BITS);
	glStencilOp(GL_KEEP, GL_ZERO, GL_ZERO);
	//keep it only where stencil test fails.  draw where stencil is true.
	//reset as we draw

	//and redraw a blue tint to it all - within the stencil buffer
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);
	glColor4f(0, 0.04f, 0.1f, 0.9f);

	Render();

	glDisable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);

	glEnable(GL_DEPTH_TEST);

	glDisable(GL_STENCIL_TEST);
}

void M_Mirror::ReflectPos(Vector *position)
{
	Vector pointOnMirror;

	plane.ClosestPoint(position, &pointOnMirror);

	position->x -= pointOnMirror.x * (1.f - reflection.x);
	position->y -= pointOnMirror.y * (1.f - reflection.y);
	position->z -= pointOnMirror.z * (1.f - reflection.z);
}
