#ifndef PARTICLE_H
#define PARTICLE_H

#include "chain.h"
#include "vector.h"
#include "types.h"
#include "camera.h"

class Particle : public Chain {
public:
	Vector		position, velocity;
	color_t		color;
	float		size;
	int			life, max_life;
	texcoord_t	blendfunc;

	float		dist;
	Particle	*child[2];	//z-sorting binary tree.  this should probably replace the chain.
};

class ParticleSystem {
public:
	ParticleSystem() {Reset();}
	~ParticleSystem() {Kill();}

	bool Init(int max);
	void Kill(void);
	void Reset(void);

	void SetBlend(int src, int dest);
	void CreateParticle(Vector *pos, Vector *vel, float size, int color, int life);

	void Update(void);
	void Display(M_Camera *camera);

private:
	Particle	*NewParticle(void);

	void		DisplayParticle(Particle *p);
	void		LinkParticle(Particle *p, Particle *node);

	Vector		view_up, view_left;

	Particle	*particle;
	int			particleCount;

	texcoord_t	last_blendfunc;

	Chain		liveChain, deadChain;
	Chain		textureChain;
};

#endif