#include "quat.h"
#include "const.h"
#include <math.h>

Quat::Quat()
{
	x = y = z = 0;
	w = 1;
}

Quat::Quat(float vx, float vy, float vz, float vw)
{
	x = vx; y = vy; z = vz; w = vw;
}

Quat::Quat(Vector *v)	//constructs it from a vector with w=0
{
	x = v->x; y = v->y; z = v->z; w = 0;
}

void Quat::QuatToAngleAxis(Quat *src)
{
	float scale, angle;

	angle = (float)acos(src->w);

	scale = (float)sin(angle);
//	scale = (float)sqrt(1.0 - src->w*src->w);	//mathematically simplified

	if (scale)	//if a scale exists, then divide away
	{
		scale = 1.f/scale;
		w = angle*360.0f/(float)M_PI;
		x = src->x*scale;
		y = src->y*scale;
		z = src->z*scale;
	}
	else		//if not then set it to 0 degrees rotation around the axis [0,0,1]
	{
		w = x = y = 0;
		z = 1;
	}

}

void Quat::AngleAxisToQuat(Quat *src)
{
	float scale, angle;

	angle = src->w*(float)M_PI/360.0f;
	scale = (float)sin(angle);

	w = (float)cos(angle);
	x = src->x * scale;
	y = src->y * scale;
	z = src->z * scale;
}

//multiplies q1 and q2, stores the result in the calling quaternion
void Quat::Mul(Quat *q1, Quat *q2)
{
	float a,b,c,d,e,f,g,h;

	a = (q1->w + q1->x)*(q2->w + q2->x);
	b = (q1->z - q1->y)*(q2->y - q2->z);
	c = (q1->x - q1->w)*(q2->y + q2->z);
	d = (q1->y + q1->z)*(q2->x - q2->w);
	e = (q1->x + q1->z)*(q2->x + q2->y);
	f = (q1->x - q1->z)*(q2->x - q2->y);
	g = (q1->w + q1->y)*(q2->w - q2->z);
	h = (q1->w - q1->y)*(q2->w + q2->z);

	w =  b + 0.5f*(-e - f + g + h);
	x =	 a - 0.5f*( e + f + g + h);
	y =	-c + 0.5f*( e - f + g - h);
	z =	-d + 0.5f*( e - f - g + h);
}

void Quat::Rotate(Vector *in, Vector *out)
{
	Quat res, conj, qin(in);

	conj = *this;
	conj.Conj();

	res.Mul(this, &qin);
	res.Mul(&res, &conj);

	*out = *(res.AxisVector());
}


void Quat::Slerp(Quat *from, Quat *to, double t)
{
	Quat to1;
	double omega, cosom, sinom, scale0, scale1;

	//a quaternion dot.  implement me if we get more requests?
	cosom = from->w*to->w + from->x*to->x + from->y*to->y + from->z*to->z;

	if (cosom < 0.f)
	{
		cosom = -cosom;
		to1.w = -to->w;
		to1.x = -to->x;
		to1.y = -to->y;
		to1.z = -to->z;
	}
	else
	{
		to1 = *to;
	}
	//therefore cosom always ranges from 0 to 1, and omega always ranges from -180 to 180 degrees

	if (cosom < .99999)	//SLERPing
	{
		omega = acos(cosom);
		sinom = sin(omega);
		scale0 = sin((1.f - t) * omega) / sinom;
		scale1 = sin(t * omega) / sinom;
	}
	else				//LERPing
	{
		scale0 = 1.f - t;
		scale1 = t;
	}

	LERP(from, &to1, (float)scale0, (float)scale1);
}

void Quat::LERP(Quat *src, Quat *dst, float src_w, float dst_w)
{
	w = src_w * src->w + dst_w * dst->w;
	x = src_w * src->x + dst_w * dst->x;
	y = src_w * src->y + dst_w * dst->y;
	z = src_w * src->z + dst_w * dst->z;
}

float Quat::Length(void)
{
	return (float)sqrt(x*x+y*y+z*z+w*w);
}
