#ifndef QUAT_H
#define QUAT_H

//{[x,   y,   z],  w}
//v[0],v[1],v[2],v[3]

#include "vector.h"

class Quat {
public:
	union {
		float v[4];
		struct {
			float x,y,z,w;
		};
	};

	//constructors
	Quat();
	Quat(float vx, float vy, float vz, float vw);
	Quat(Vector *v);	//constructs it from a vector with w=0

	//transformers (ROBOTS IN DISGUISE!!!)
	void QuatToAngleAxis(Quat *src);
	void QuatToAngleAxis() {QuatToAngleAxis(this);}

	void AngleAxisToQuat(Quat *src);
	void AngleAxisToQuat() {AngleAxisToQuat(this);}

	//mathematical functions
	void Mul(Quat *q1, Quat *q2);	//multiplies q1 and q2, stores the result in the calling quaternion
	void MulTo(Quat *q) {Mul(this,q);}	//the equiv of this = this * q (as opposed to this = q * this)
	void operator*=(Quat &q) {Mul(this,&q);}

	void Conj(void) {x = -x; y = -y; z = -z;}

	//rotation
	void Rotate(Vector *in, Vector *out);
	void Rotate(Vector *v) {Rotate(v,v);}

	//conversion
	Vector *AxisVector() {return (Vector *)&x;}

	//spherical linear interpolation
	void Slerp(Quat *from, Quat *to, double t);

	//the source quat, the dest quat, and the appropriate weights (naturally 
	void LERP(Quat *src, Quat *dst, float src_w, float dst_w);
	void LERP(Quat *src, Quat *dst, float coeff) {LERP(src, dst, 1.f - coeff, coeff);}

	void Scale(float v) {x*=v;y*=v;z*=v;w*=v;}
	float Length(void);
	void Normalize(void) {Scale(1.f/Length());}

};

#endif