#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>

#include "texture.h"
#include "error.h"

#include "main.h"	//main_console

// ********** CONSTRUCTION/DESTRUCTION ********** 

//create an Texture
Texture::Texture()
{
	texture_id = 0;

	width = height = 0;
	data = NULL;
}

void Texture::Destroy(void)
{
	if (texture_id) glDeleteTextures(1, &texture_id);
	Unlink();
}

// ********** SETTING CURRENT TEXTURE ********** 

void SetTexture(Texture *t)
{
	int id = 0;
	if (t) id = t->texture_id;
	glBindTexture(GL_TEXTURE_2D, id);
}

// ********** TEXTURE PRE-CREATION ********** 

#include "console.h"

bool Texture::CreateEmpty(int bytes_per_pixel, int width, int height)
{
	data = (unsigned char *)malloc(width * height * bytes_per_pixel);
	if (!data) return false;

	memset(data, 0, width * height * bytes_per_pixel);

	glEnable(GL_TEXTURE);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	//the per-texture stuff:
	glGenTextures(1, &texture_id);

	glEnable(GL_TEXTURE_2D);

	glBindTexture(GL_TEXTURE_2D, texture_id );

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);

	int format = GL_RGB;
	if (bytes_per_pixel == 1) format = GL_ALPHA;
	else if (bytes_per_pixel == 3) format = GL_RGB;
	else if (bytes_per_pixel == 4) format = GL_RGBA;

	glTexImage2D(GL_TEXTURE_2D, 0, bytes_per_pixel, width, height, 0, format, GL_UNSIGNED_BYTE, data);

// --- FREE --- 
	free(data);	data = NULL;

	main_console->Print("sucessfully extracted texture %d\n", texture_id);

	return true;
}

void Texture::UpdateData(int bytes_per_pixel, int width, int height, void *src)
{
	int format = GL_RGB;
	if (bytes_per_pixel == 3) format = GL_RGB;
	else if (bytes_per_pixel == 4) format = GL_RGBA;

	glBindTexture(GL_TEXTURE_2D, texture_id);
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height, format, GL_UNSIGNED_BYTE, src);
}

// ********** TEXTURE ARRAY EXTRACTION ********** 
bool Texture::Extract(unsigned char *ptr, int bytes_pp, int width, int height, int colskip, int rowskip)
{
	data = (unsigned char *)malloc(width * height * bytes_pp);
	if (!data) return false;

	int i,j;
	unsigned char *psrc_x, *psrc_y, *pdest;

	pdest = data;
	psrc_y = ptr;
	for (j = 0; j < height; j++)
	{
		psrc_x = psrc_y;
		for (i = 0; i < width; i++)
		{
			memcpy(pdest, psrc_x, bytes_pp);
			pdest += bytes_pp;
			psrc_x += colskip;
		}
		psrc_y += rowskip;
	}

	glEnable(GL_TEXTURE);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glGenTextures(1, &texture_id);
	glEnable(GL_TEXTURE_2D);

	glBindTexture(GL_TEXTURE_2D, texture_id);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);

	bytes_per_pixel = bytes_pp;

	int format = GL_RGB;
	if (bytes_per_pixel == 1) format = GL_LUMINANCE;
	if (bytes_per_pixel == 2) format = GL_LUMINANCE_ALPHA;
	if (bytes_per_pixel == 3) format = GL_RGB;
	if (bytes_per_pixel == 4) format = GL_RGBA;

	gluBuild2DMipmaps (GL_TEXTURE_2D, bytes_pp, width, height, format, GL_UNSIGNED_BYTE, data);

// --- FREE --- 
	free(data);	data = NULL;

	main_console->Print("sucessfully extracted texture %d\n", texture_id);

	return true;
}

// ********** TEXTURE IMAGE LOADING ********** 

		//load the texture - return the texture ID
bool Texture::Create(char *fn)
{
	strcpy(title, fn);

// --- ALLOCATE AND LOAD --- 
	//attempt to load the texture data *here*
	if (!LoadData(fn)) return false;

// --- SET OPENGL VARS --- 
	//globals - only needed to be called once
	glEnable(GL_TEXTURE);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	//the per-texture stuff:
	glGenTextures(1, &texture_id);

	glEnable(GL_TEXTURE_2D);
	
	glBindTexture(GL_TEXTURE_2D, texture_id );

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);

// --- LOAD TEXTURE --- 

	//load the texture data into Video RAM
/*
	glTexImage2D(
		GL_TEXTURE_2D,		//gl_texture_2d
		0,					//mipmap level
		3,					//internal format
		width,		//texture width
		height,	//texture height
		0,					//use border?
		GL_RGB,				//pixel data format
		GL_UNSIGNED_BYTE,	//pixel data type
		data);		//pixel data pointer
*/
	int format = GL_RGB;
	if (bytes_per_pixel == 3) format = GL_RGB;
	else if (bytes_per_pixel == 4) format = GL_RGBA;

	gluBuild2DMipmaps (GL_TEXTURE_2D, bytes_per_pixel, width, height, format, GL_UNSIGNED_BYTE, data);

// --- FREE --- 
	free(data);	data = NULL;

	main_console->Print("sucessfully loaded texture %s as %d\n", fn, texture_id);

	return true;
}

// *********** LOAD TEXTURE *********
// responcible for choosing which of the above functions to use
// in loading data

/*
LoadTexture()
loads the texture image from the file specified in "filename"
and stores the contents in the texture_t structure
returns true on sucess, or false on failure
*/
bool Texture::LoadData(char *fn)
{
	//set our method vars
	char ext[4] = {0};
	int i = strlen(fn);

	//initialize texture image vars
	width = height = 0;
	data = NULL;

	//copy over the file extension
	memcpy(ext, &fn[i-3], 3);

	//properly case the file extension
	if (ext[0] >= 'A' && ext[0] <= 'Z') ext[0] += 'a' - 'A';
	if (ext[1] >= 'A' && ext[1] <= 'Z') ext[1] += 'a' - 'A';
	if (ext[2] >= 'A' && ext[2] <= 'Z') ext[2] += 'a' - 'A';

	//check for what it is
	if (memcmp(ext, "bmp", 3)==0)
	{
		if (!LoadBMP(fn)) return false;
	}
	else if (memcmp(ext, "tga", 3)==0)
	{
		if (!LoadTGA(fn)) return false;
	}
	else
	{
		main_console->Print("I found an unknown texture extension %s\n", ext);
		return false;	//unknown format
	}

	//make sure the data does exist
	if (data == NULL)
	{
		Error( "Texture::LoadData - this shouldnt have happened.");
		return false;
	}

	//convert the data based on the load flags
	ConvertData();

	//and return a big true
	return true;
}


#define TEXTURE_FLAG_BGR	0x1
#define TEXTURE_FLAG_FLIP	0x2


bool Texture::ConvertData(void)
{
	int x, y, y_step, y_end;
	unsigned char *dest, *src_ptr, *dest_ptr;

	if (!flags) return true;	//no conversion needed

	if (!data) return false;	//no data to convert

	//couldnt allocate temp data
	if (!(dest = (unsigned char *)malloc(width * height * bytes_per_pixel))) return false;


	if (flags & TEXTURE_FLAG_FLIP)
	{
		y = height-1;
		y_step = -1;
		y_end = 0;
	}
	else
	{
		y = 0;
		y_step = 1;
		y_end = height-1;
	}

	dest_ptr = dest;
	for (; y != y_end+y_step; y += y_step)
	{
		src_ptr = &data[y*width*bytes_per_pixel];
		for (x = 0; x < (int)width; x++)
		{
			if (TEXTURE_FLAG_BGR)
			{
				dest_ptr[0] = src_ptr[2];
				dest_ptr[2] = src_ptr[0];
			}
			else
			{
				dest_ptr[0] = src_ptr[0];
				dest_ptr[2] = src_ptr[2];
			}
			dest_ptr[1] = src_ptr[1];

			if (bytes_per_pixel == 4) dest_ptr[3] = src_ptr[3];

			dest_ptr += bytes_per_pixel;
			src_ptr += bytes_per_pixel;
		}
	}

	//free original - and set 'data' to new data
	free(data);	data = dest;

	return true;
}



/*
bool TextureLoad___ (char *filename)

prereq: none
proreq:
	width contains the image in filename's width
	height contains the image in filename's height
	bytes_per_pixel contains the image in filename's bytes per pixel (3 or 4)
	data contains the interleaved RGB data

returns:
	true if loading all was sucessful
	false if an error occured (mind you, data will be freed if this occurs)
*/


//*********************************** LOAD THE IMAGE *********************************** 

/*
typedef struct {
  unsigned short        id PACKED;          //'BM' 0x4d42

  0  unsigned long         filesize PACKED;
  1  unsigned long         reserved PACKED;         //0
  2  unsigned long         pictureoffset PACKED;

  3  unsigned long         headersize PACKED;     //40
  4  unsigned long         width PACKED;
  5  unsigned long         height PACKED;
  6  unsigned short        planes PACKED;
     unsigned short        bpp PACKED;
  7  unsigned long         compression PACKED;
  8  unsigned long         picturesize PACKED;
  9  unsigned long         width_in_ppm PACKED;
  A  unsigned long         height_in_ppm PACKED;
  B  unsigned long         colors_used PACKED;
  C  unsigned long         important_colors PACKED;
} bmp_header;
*/

bool Texture::LoadBMP(char *filename)			// Loads A TGA File Into Memory
{    
	GLushort	BMPsignature = 0x4D42;
	GLushort	BMPcompare;
	GLuint		header[13];

	GLuint		bitsPerPixel;			//the src iamge's bits per pixel
	GLuint		imageSize;				//the dest image's size

	GLuint		rowSize, rowOffset;		//used for reading rows
	GLuint		tmpVar;

	unsigned char *rowPtr;

	GLubyte		palette[1024];
	GLuint		palIndexCount;

	DWORD		dwBytesRead;	//for ReadFile()


	main_console->Print("Loading Bitmap %s\n", filename);

	//open the file
	HANDLE hFile = CreateFile(filename, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE)
	{
		Error("%s: I Couldn't Open the BMP", filename);
		return false;
	}


	if ((!ReadFile(hFile, &BMPcompare, sizeof(BMPcompare), &dwBytesRead, NULL)) ||
		(dwBytesRead != sizeof(BMPcompare)))
	{
		CloseHandle(hFile);
		Error( "%s: I Couldn't Read The Bitmap Signature.  This Bitmap Is Probably Corrupt", filename);
		return false;
	}

	if (memcmp(&BMPcompare,&BMPsignature,sizeof(GLushort))!=0)
	{
		CloseHandle(hFile);
		Error("s: The Bitmap ID Was Incorrect.  This Bitmap Is Probably Corrupt", filename);
		return false;
	}

	if ((!ReadFile(hFile, header, sizeof(header), &dwBytesRead, NULL)) ||
		(dwBytesRead != sizeof(header)))

	{
		CloseHandle(hFile);
		Error("s: I Couldnt Read The Bitmap Header.  This Bitmap Is Probably Corrupt", filename);
		return false;
	}

	width  = header[4];
	height = header[5];
	bitsPerPixel = HIWORD(header[6]);
	palIndexCount = header[11];

	if (bitsPerPixel != 1 && 
		bitsPerPixel != 4 &&
		bitsPerPixel != 8 &&
		bitsPerPixel != 24)
	{
		CloseHandle(hFile);
		Error("s: The Bits Per Pixel Is %d, An Unsupported Value.", filename, bitsPerPixel);
		return false;
	}

	if (header[7])
	{
		CloseHandle(hFile);
		Error("s: This Screensaver Does Not Yet Support RLE Compressed BMP files.  This Image is RLE Compressed", filename);
		return false;
	}

	if (bitsPerPixel <= 8)	bytes_per_pixel = 3;
	else					bytes_per_pixel = (bitsPerPixel>>3);

	//for allocation...
	imageSize	= width * height * bytes_per_pixel;

	data = (unsigned char *)malloc(imageSize);

	if(data==NULL)
	{
		Error("s: I Couldnt Allocate Memory To Load The Image", filename);
		CloseHandle(hFile);
		return false;
	}

	if (tmpVar = (header[2] - 54))					//if an offset to the image exists...
	{
		unsigned int palSize, extraSize;
		//if (palIndexCount > 256) then we cant grab all the indecies

		if (palIndexCount == 0) palSize = tmpVar > 1024 ? 1024 : tmpVar;
		else palSize = (palIndexCount<<2);

		extraSize = tmpVar - palSize;

		if (palSize > 0)
		{

			//if (fread(palette, 1, palSize, file) != palSize)
			if ((!ReadFile(hFile, palette, palSize, &dwBytesRead, NULL)) ||
				(dwBytesRead != palSize))
			{
				free(data);	data = NULL;
				CloseHandle(hFile);
				Error("s: There Was An Error Reading The Palette", filename);
				return false;
			}
		}

		if (extraSize > 0)
		{
//			if (fseek(file, extraSize, SEEK_CUR))
			SetFilePointer(hFile, extraSize, NULL, FILE_CURRENT);
			if (GetLastError() != NO_ERROR)
			{
				free(data);	data = NULL;
				CloseHandle(hFile);
				Error("s: There Was An Error Seeking To The Image Data", filename);
				return false;
			}
		}
	}

	//for reading the rows of the bitmap and for calc'ing the row offset
	rowSize = width * bitsPerPixel;
	rowSize = (rowSize>>3) + !!(rowSize & 7);

	//waste of space - thank the people at microsoft for this idea - the row offset
	rowOffset		= rowSize&3;
	if (rowOffset)	rowOffset = 4-rowOffset;

	//used for indexed bitmaps:
	GLubyte curByte;
	int current_bit_offset = 8 - bitsPerPixel;
	int pixels_per_byte = 8 / bitsPerPixel;
	int pixel_bitmask = (1 << bitsPerPixel) - 1;
	int index;

	rowPtr = data;
	for (int y = 0; y < (int)height; y++)
	{
		//load the poop as indexed, then convert it from the palette
		if (bitsPerPixel <= 8)
		{
			//load these per-byte and convert them as we go
			for (int x = 0; x < (int)width; x++)
			{
				if ((x % pixels_per_byte) == 0)
				{
					current_bit_offset = 8 - bitsPerPixel;
					if ((!ReadFile(hFile, &curByte, 1, &dwBytesRead, NULL)) ||
						(dwBytesRead != 1))
					{
						free(data);	data = NULL;
						CloseHandle(hFile);
						Error("s: There Was An Error Reading Image Byte <%d,%d>", filename, x, y);
						return false;
					}

					index = (curByte >> current_bit_offset) & pixel_bitmask;
					current_bit_offset -= bitsPerPixel;
					memcpy(rowPtr, &palette[index<<2], bytes_per_pixel);
				}
				else
				{
					index = (curByte >> current_bit_offset) & pixel_bitmask;
					current_bit_offset -= bitsPerPixel;
					memcpy(rowPtr, &palette[index<<2], bytes_per_pixel);
				}
				rowPtr += bytes_per_pixel;
			}
		}
		else
		{	//load the poop as true color
			if ((!ReadFile(hFile, rowPtr, rowSize, &dwBytesRead, NULL)) ||
				(dwBytesRead != rowSize))
			{
				free(data);	data = NULL;
				CloseHandle(hFile);
				Error("s: There Was An Error Reading Image Row %d", filename, y);
				return false;
			}
			rowPtr += rowSize;
		}

		if (rowOffset)
		{
			SetFilePointer(hFile, rowOffset, NULL, FILE_CURRENT);
			if (GetLastError() != NO_ERROR)
			{
				free(data);	data = NULL;
				CloseHandle(hFile);
				Error("s: I was Unable to Seek Through the Offset of Row %d", filename, y);
				return false;
			}
		}
	}

	CloseHandle(hFile);

	flags |= TEXTURE_FLAG_BGR | TEXTURE_FLAG_FLIP;

	main_console->Print("Loaded Bitmap %s\n", filename);

	return true;
}

	/*
	TGA header:
	{
		0	1 byte		imageID			-> number of bytes in colormap first index
		1	1 byte		colorMapType
		2	1 byte		imageType

		combinations of [1]::[2]
		{
			1,1 -> color mapped
			0,2 -> true color
			0,3 -> greyscale
			1,9 -> RLE encoded color mapped
			0,10 -> RLE encoded true color
			0,11 -> RLE greyscale
		}

						(zero if no colormap is used):
		3	2 short	colorMap.firstEntryIndex - starting entry of the colormap
		5	2 short	colorMap.length  - number of entries in the color map
		7	1 byte	colorMap.entrySize - the number of bits per entry: (15,16,24,32)

		8	2 short		x origin of image
		10	2 short		y origin of image
		12	2 short		width of image
		14	2 short		height of image
		16	1 byte		bits per pixel
		17	1 byte		image descriptor
	}

	next comes the colormap info - size equal to colorMap.length * bytes per each colorMap entry

    and then the BGR interleaved image data
	*/


bool Texture::LoadTGA(char *filename)
{
	GLubyte		header[18];
	GLuint		bitsPerPixel;
	GLuint		imageSize;
	GLuint		rowSize;

	unsigned char *rowPtr;


	main_console->Print("Loading TGA %s\n", filename);

	DWORD dwBytesRead;

	//open the file
	HANDLE hFile = CreateFile(filename, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE)
	{
		Error("s: unable to open file", filename);
		return false;
	}

//	if (fread(header,1,sizeof(header),file) != sizeof(header))
	if ((!ReadFile(hFile, header, sizeof(header), &dwBytesRead, NULL)) ||
		(dwBytesRead != sizeof(header)))
	{
		CloseHandle(hFile);
		Error("couldnt read header");
		return false;
	}

//	if (fseek(file, header[0], SEEK_CUR))
	SetFilePointer(hFile, header[0], NULL, FILE_CURRENT);
	if (GetLastError() != NO_ERROR)
	{
		CloseHandle(hFile);
		Error("error trying to seek past image ID info");
		return false;
	}


	//indexed color stuff:
	unsigned char *colorMap = NULL;
	int colorMapCount;
	int colorMapBits, colorMapBytes;

	if (header[1])		//a color map is present - lets read it
	{
		colorMapCount = *(GLushort *)(&header[5]);
		colorMapBits = header[7];

		//my way of rounding up - base 8
		colorMapBytes = (colorMapBits>>3) + !!(colorMapBits & 7);

		colorMap = (unsigned char *)malloc(colorMapCount * colorMapBytes);
		if (colorMap == NULL)
		{
			Error("couldnt allocate memory for the TGA's palette");
			CloseHandle(hFile);
			return false;
		}
	}

	int imageType = header[2] & 7;
	int RLEcompressed = header[2] & 8;
	if (RLEcompressed)
	{
		Error("this TGA file just happens to be RLE, which I dont support yet");
		if (colorMap) free(colorMap);
		colorMap = NULL;
		CloseHandle(hFile);
		return false;
	}

	width = *(GLushort *)(&header[12]);
	height = *(GLushort *)(&header[14]);
	bitsPerPixel = header[16];
	bytes_per_pixel = bitsPerPixel>>3;

	if (header[17] & 0x20) flags |= TEXTURE_FLAG_FLIP;		//vertical flip
	//if (header[17] & 0x10) flags |= TEXTURE_FLAG_MIRROR;		//horizontal flip

 	if(bitsPerPixel != 8 && bitsPerPixel !=24 && bitsPerPixel!=32)
	{
		if (colorMap) free(colorMap);
		colorMap = NULL;
		CloseHandle(hFile);
		Error("TGA bpp != 8,24,32");
		return false;
	}

	imageSize = width * height * bytes_per_pixel;

	data = (unsigned char *)malloc(imageSize);
	if(data==NULL)
	{
		CloseHandle(hFile);
		if (colorMap) free(colorMap);
		colorMap = NULL;
		Error("couldnt allocate the image data");
		return false;
	}

//	for(temp=0; temp<(int)imageSize; temp++) data[temp]=0;

	rowSize	= width * bytes_per_pixel;

	rowPtr = data;
	int x;
	for (int y = 0; y < (int)height; y++)
	{
		if (bitsPerPixel == 8)
		{
			GLubyte t;
			for (x = 0; x < (int)width; x++)
			{
//				if (fread(&t, 1, 1, file) != 1)
				if ((!ReadFile(hFile, &t, 1, &dwBytesRead, NULL)) ||
					(dwBytesRead != 1))
				{
					if (colorMap) free(colorMap);
					colorMap = NULL;
					free(data);	data = NULL;
					CloseHandle(hFile);
					Error("error in reading this TGA");
					return false;
				}

				if (header[1] == 0 && header[2] == 3)	//greyscale 8bit image
				{
					rowPtr[2] = rowPtr[1] = rowPtr[0] = t;
				}
				else if (header[1] == 1 && header[2] == 1)
				{
					if (colorMap == NULL)
					{
						Error("colormap hasnt been found");
						free(data);	data = NULL;
						CloseHandle(hFile);
						return false;
					}
					memcpy(rowPtr, &colorMap[t*3], 3);
				}

				rowPtr += 3;
			}
		}
		else
		{
//			if (fread(rowPtr, 1, rowSize, file) != rowSize)
			if ((!ReadFile(hFile, rowPtr, rowSize, &dwBytesRead, NULL)) ||
				(dwBytesRead != rowSize))
			{
				free(data);	data = NULL;
				if (colorMap) free(colorMap);
				colorMap = NULL;
				CloseHandle(hFile);
				Error("couldnt read texture row");
				return false;
			}
			rowPtr += rowSize;
		}
	}

	if (colorMap) free(colorMap);
	colorMap = NULL;

	CloseHandle (hFile);

	//swap all the Bs and Rs
	flags |= TEXTURE_FLAG_BGR;

	main_console->Print("Loaded TGA %s\n", filename);

	return true;
}

Texture *GetTexture(Chain *p, char *title)
{
	Texture *t = (Texture *)p->ChainHead();
	for (; t; t = (Texture *)t->ChainNext())
		if (!_stricmp(title, t->title)) break;

	return t;
}
