#ifndef VECTOR_H
#define VECTOR_H

//  x    y    z
//v[0] v[1] v[2]

class Vector {
public:
	union {
		float v[3];
		struct {
			float x,y,z;
		};
	};

	Vector();
	Vector(float vx, float vy, float vz);
	Vector(Vector *v)	{x = v->x; y = v->y; z = v->z;}

	void add(Vector *v) {x += v->x; y += v->y; z += v->z;}
	void add(Vector *a, Vector *b) {x = a->x + b->x; y = a->y + b->y; z = a->z + b->z; }
	void operator+=(const Vector &a) {x+=a.x;y+=a.y;z+=a.z;}

	void sub(Vector *v) {x -= v->x; y -= v->y; z -= v->z;}
	void sub(Vector *a, Vector *b) {x = a->x - b->x; y = a->y - b->y; z = a->z - b->z; }
	void scale(float f)	{x *= f; y *= f; z *= f;}
	void scale(Vector *a, float f) {x = a->x * f; y = a->y * f; z = a->z * f;}
	float dot(Vector *a) { return (a->x*x + a->y*y + a->z*z); }
	void cross(Vector *a, Vector *b) {
		x = a->y*b->z - a->z*b->y;
		y = a->z*b->x - a->x*b->z;
		z = a->x*b->y - a->y*b->x;
	}

	void negative(void) {x=-x;y=-y;z=-z;}

	float normal(void) {return (x*x + y*y + z*z);}
	float length(void);
	void normalize(void) {scale(1.f/length());}

	void LERP(Vector *start, Vector *end, float coeff);

	//calc the normal of the plane based upon the three vectors
	//and store it in the current vector
	void CalcNormal(Vector *v1, Vector *v2, Vector *v3);
};

extern Vector vectorInf;
extern Vector base_axis[3];

#endif