#include <windows.h>
#include <gl/gl.h>
#include <math.h>

#include "world.h"
#include "vector.h"
#include "geometry.h"
#include "entity.h"
#include "fractal.h"

M_World::M_World()
{
	worldtex = NULL;
}

bool M_World::Init(void)
{
	float *tptr;
	unsigned char *texptr;

	texptr = (unsigned char *)malloc(256 * 256 * 3);
	if (!texptr) return true;

	tptr = (float *)malloc(256 * 256 * sizeof(float));
	if (!tptr)
	{
		free (texptr);
		return true;
	}

	Fractal_Create(tptr, 0, 255, 256, 256, 4, 0);
	for (int i = 0; i < 3; i++)
	{
		for (int v = 0; v < 65536; v++)
		{
			texptr[3*v+i] = (unsigned char)tptr[v];
		}
	}

	free(tptr);

	worldtex = new Texture();
	if (!worldtex)
	{
		free(texptr);
		return true;
	}

	worldtex->Extract(texptr, 3, 256, 256, 3, 768);
	
	return true;
}

void M_World::Kill(void)
{
	if (worldtex) delete worldtex;
	worldtex = NULL;
}

void DrawLaserSight(void);

#define SQ13	((float)0.577350269189625731058868041145615279675)

void M_World::Display(void)
{

	Box worldBox;
	Vector corner[8];

	static int _lines[24] = {0,1,1,3,3,2,2,0,4,5,5,7,7,6,6,4,0,4,1,5,2,6,3,7};
	static int planes[24] = {0,1,3,2,4,5,7,6,0,1,5,4,1,5,7,3,2,3,7,6,2,0,4,6};
	static float normals[8][3] = {
		{SQ13,SQ13,SQ13},	{SQ13,SQ13,-SQ13},	{SQ13,-SQ13,SQ13},	{SQ13,-SQ13,-SQ13},
		{-SQ13,SQ13,SQ13},	{-SQ13,SQ13,-SQ13},	{-SQ13,-SQ13,SQ13},	{-SQ13,-SQ13,-SQ13}};

	worldBox.min = Vector(-WorldSize,-WorldSize,-WorldSize);
	worldBox.max = Vector(WorldSize,WorldSize,WorldSize);
	worldBox.CalcVtxs(corner);

	glColor4f(.5f,.5f,.5f,.5f);
	SetTexture(worldtex);

glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);

glEnable(GL_TEXTURE_GEN_S);
glEnable(GL_TEXTURE_GEN_T);

	glPushAttrib(GL_ALL_ATTRIB_BITS);
/*	glDisable(GL_CULL_FACE);

	glBegin(GL_LINES);
	for (int v = 0; v < 24; v++) glVertex3fv(corner[_lines[v]].v);
	glEnd();

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	glDepthMask(GL_FALSE);
	glBegin(GL_QUADS);
	for (v = 0; v < 24; v++)
	{
		glNormal3fv(normals[planes[v]]);
		glVertex3fv(corner[planes[v]].v);
	}
	glEnd();

	glDepthMask(GL_TRUE);
	glDisable(GL_BLEND);
*/
	Vector vec;

#define STEP_INC	1000


	glBindTexture(GL_TEXTURE_2D, 0);
	glColor4f(.5f,.5f,.5f,.5f);
	glBegin(GL_LINES);
	for (vec.x = -WorldSize; vec.x < WorldSize+1; vec.x += STEP_INC)
	{
		for (vec.y = -WorldSize; vec.y < WorldSize+1; vec.y += STEP_INC)
		{
			for (vec.z = -WorldSize; vec.z < WorldSize+1; vec.z += STEP_INC)
			{
				if (vec.x < WorldSize-1)
				{
					glNormal3f(1,0,0);
					glVertex3fv(vec.v);
					glVertex3f(vec.x + STEP_INC, vec.y, vec.z);
				}
				if (vec.y < WorldSize-1)
				{
					glNormal3f(0,1,0);
					glVertex3fv(vec.v);
					glVertex3f(vec.x, vec.y + STEP_INC, vec.z);
				}
				if (vec.z < WorldSize-1)
				{
					glNormal3f(0,0,1);
					glVertex3fv(vec.v);
					glVertex3f(vec.x, vec.y, vec.z + STEP_INC);
				}
			}
		}
	}
	glEnd();

	//main client display (space is the place targetting laser)
	DrawLaserSight();

	glPopAttrib();
}

#include "const.h"
void DrawLaserSight(void)
{
	Vector z_dir, x_dir, y_dir, src, dest;
	Vector tx,ty,tz;
	float theta;

	if (!client_ent) return;

//fix this poop up - project along 3 axii

	tx.scale(&client_ent->axis[0], 5000);			//scale it to the end

	static float tex_offset = 0;

	glColor3f(1,0,0);
	for (theta = 0; theta < 2.f*(float)M_PI; theta += (float)M_PI*0.5f)
	{
		ty.scale(&client_ent->axis[1], 25.f*(float)cos(theta + tex_offset));
		tz.scale(&client_ent->axis[2], 25.f*(float)sin(theta + tex_offset));

		src.add(&client_ent->position, &ty);
		src.add(&tz);
		dest.add(&src, &tx);	//z_dir is scaled

		glBegin(GL_LINES);
		glTexCoord2f(tex_offset + theta,0); glVertex3fv(src.v);
		glTexCoord2f(tex_offset + theta,1); glVertex3fv(dest.v);
		glEnd();
	}

	tex_offset += 1.f/32.f;
}
