#include <windows.h>

#include "display.h"
#include "engine.h"
#include "error.h"
#include "console.h"

#include "entity.h"	//debug

M_Display *main_display = NULL;
M_Engine *main_engine = NULL;
M_KeyboardHandler *main_keyboard = NULL;
M_Console *main_console = NULL;

//put me somewhere
#include <math.h>
double Inf = -log(0.0);
float fInf = -(float)log(0.0);

int DisplayFPS = 0;
float CurrentFPS;
void MainInitCVars(void)
{
	static bool init_d = false;
	if (!init_d)
	{
		init_d = true;
		new M_CVar(&DisplayFPS, "DisplayFPS");
		new M_CVar(&CurrentFPS, "CurrentFPS");
	}
}

bool Init_System(void)
{
	//create the keyboard handler
	main_keyboard = new M_KeyboardHandler();

	//create the display - window and OpenGL context
	main_display = new M_Display();
	if (main_display && !main_display->Init())
	{
		delete main_display;
		main_display = NULL;
	}
	if (!main_display)
	{
		Error("Failed to create display");
		return false;
	}

	main_display->Clear();

	//what does display have to do with the keyboard?
	//well, since im using the WM_KEYUP/DOWN messages to get keyboard input
	//input is going to be trapped via window
	//and main_display contains the main display window.
	//so there.
	//of course - this means we have to kill the keyboard handler only after we kill the window
	main_display->SetKeyboardHandler(main_keyboard);

	return true;
}

bool Init(void)
{
	//create the main console variables
	MainInitCVars();

	//init the hardware-related components
	if (!Init_System()) return false;

//get the console up ASAP - right after display
	main_console = new M_Console();
	if (main_console) main_console->Init(80,50);

	//create the engine
	main_engine = new M_Engine();
	if (main_engine && !main_engine->Init())
	{
		delete main_engine;
		main_engine = NULL;
	}
	if (!main_engine)
	{
		Error("Failed to create engine");
		return false;
	}

	//now that the engine is up - splash screen??? help???

	return true;
}

void Update(void)
{
	//calculate the framerate
	{
		static int lasttime=-1,thistime, framecount=0;
		framecount++;

		if ((lasttime/1000) != ((thistime=timeGetTime())/1000))
		{
			CurrentFPS = (float)framecount * 1000.f / (float)(thistime - lasttime);
			if (DisplayFPS) main_console->Print("FPS: %f\n", CurrentFPS);
			lasttime = thistime;
			framecount = 0;
		}
	}

	if (main_keyboard) main_keyboard->Update();

	if (main_engine) main_engine->Update();

	if (main_console) main_console->Display();

	if (main_display) main_display->Update();
}

void Shutdown(void)
{
	if (main_console)	{delete main_console;	main_console = NULL;	}
	if (main_engine)	{delete main_engine;	main_engine = NULL;		}
	if (main_display)	{delete main_display;	main_display = NULL;	}
	if (main_keyboard)	{delete main_keyboard;	main_keyboard = NULL;	}
}

//the one and only:
int APIENTRY WinMain(HINSTANCE hCurrentInst, HINSTANCE hPreviousInst, LPSTR lpszCmdLine, int nCmdShow)
{
    MSG msg;				/* message */
	HWND hWnd = NULL;

	if (!Init())
	{
		Shutdown();
		return 1;
	}

//keep this puppy NULL so we can update both the console and the main window
	//	if (main_display) hWnd = main_display->GetHWnd();

	while (1) {
		while(PeekMessage(&msg, hWnd, 0, 0, PM_NOREMOVE)) {
			if(GetMessage(&msg, hWnd, 0, 0)) {
				TranslateMessage(&msg);
				DispatchMessage(&msg);
		    } else {
				goto quit; // This 'goto' was in the sample code
			}
		}

		Update();
	}

quit:
	Shutdown();

    return msg.wParam;
}
