#include "oglh.h"
#include "error.h"

#include <windows.h>
#include <gl/gl.h>

#include "console.h"	//i am probably going to get in trouble for this

// *************** OPENGL HANDLER CONSOLE VAR STUFF ***************

void CV_OpenGLHandler_Update(int argc, char **argv, void *param)
{
	OpenGLHandler *owner = (OpenGLHandler *)param;
	if (!owner) return;
	owner->CV_UpdateVars();
}

void OpenGLHandler::CV_UpdateVars(void)
{
//update PolygonMode
	int true_PolygonMode;
	switch (PolygonMode) {
	case (0):	true_PolygonMode = GL_POINT; break;
	case (1):	true_PolygonMode = GL_LINE; break;
	default:	true_PolygonMode = GL_FILL; break;
	}
	glPolygonMode(GL_FRONT_AND_BACK, true_PolygonMode);
}

// ************************* CONSTRUCTOR  *************************

OpenGLHandler::OpenGLHandler()
{
	active = false;
	hDC = NULL;
	hRC = NULL;
	hWindow = NULL;
	doublebuffered = false;

	width = 1;
	height = 1;

	PolygonMode = 2;	//fill 

	new M_CVar(&PolygonMode,"PolygonMode",	CV_OpenGLHandler_Update, this);
}

OpenGLHandler::~OpenGLHandler() {Kill();}

// ******************************* OPENGL HANDLER *******************************

bool OpenGLHandler::Init(HWND hWnd, bool dbuf)
{
	PIXELFORMATDESCRIPTOR pfd;
	int pf;

	hWindow = hWnd;
	doublebuffered = dbuf;

	if ((hDC = GetDC(hWindow))==NULL)
	{
		Error("OpenGLHandler::Init() : GetDC() failed");
		return false;
	}

	//clear the structure
	memset(&pfd, 0, sizeof(pfd));

	//get the current pixel format
	int iPixelFormat = GetPixelFormat(hDC); 

	//and get its description
	DescribePixelFormat(hDC, iPixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &pfd); 

	//modify its description accordingly:
	pfd.nSize		= sizeof(pfd);
	pfd.nVersion	= 1;
	
	pfd.dwFlags		= PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL;
	if (doublebuffered) pfd.dwFlags |= PFD_DOUBLEBUFFER;

	pfd.iPixelType	= PFD_TYPE_RGBA;
	pfd.iLayerType	= PFD_MAIN_PLANE;


	//FIXME: stencil RW stutters like mad when run in 16 bit color - not HW accelerated? whats up?
	//pfd.cColorBits = 32;

	pfd.cStencilBits = 16;

	//find the best fit pixel format to set it to
	if ((pf = ChoosePixelFormat(hDC, &pfd)) == 0)
	{
		Error("OpenGLHandler::Init() : ChoosePixelFormat() failed");
		return false;
	}

	//and set it accordingly
	if (SetPixelFormat(hDC, pf, &pfd) == false)
	{
		Error("OpenGLHandler::Init() : SetPixelFormat() failed");
		return false;
	}


	//make the GL RC
	if ((hRC = wglCreateContext(hDC))==NULL)
	{
		Error("OpenGLHandler::Init() : wglCreateContext() failed");
		return false;
	}

	//and make it the current RC
	if (wglMakeCurrent(hDC, hRC)==false)
	{
		Error("OpenGLHandler::Init() : wglMakeCurrent() failed");
		return false;
	}

	active = true;

	Resize();

	return true;
}

void OpenGLHandler::Kill(void)
{
	if (!active) return;

	wglMakeCurrent(NULL, NULL);
	ReleaseDC(hWindow, hDC);
	wglDeleteContext(hRC);

	active = false;
}

void OpenGLHandler::Clear(void) { glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT); }

void OpenGLHandler::Update(void)
{
	if (!active) return;
	if (doublebuffered) SwapBuffers(hDC);
	else 				glFlush();
}

void OpenGLHandler::Resize(void)
{
	RECT winRect;

	if (!active) return;

	if (!GetClientRect(hWindow, &winRect)) return;
	
	width = winRect.right;
	height = winRect.bottom;
	glViewport(0, 0, width, height);
}

int OpenGLHandler::Width(void)	{return width;}
int OpenGLHandler::Height(void)	{return height;}
